/* 
 * Copyright (C) 2005  Network Applied Communication Laboratory Co., Ltd.
 *
 * This file is part of Rast.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA
 *
 * Linking the Rast statically or dynamically with other modules is
 * making a combined work based on the Rast.  Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * As a special exception, the copyright holders of the Rast give
 * you permission to link the Rast with independent modules that
 * communicate with the Rast solely through the the Apache Portable
 * Runtime interface, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * combined work under terms of your choice, provided that every
 * copy of the combined work is accompanied by a complete copy of
 * the source code of the Rast (the version of the Rast used to
 * produce the combined work), being distributed under the terms of
 * the GNU General Public License plus this exception.  An
 * independent module is a module which is not derived from or based
 * on the Rast.
 *
 * Note that people who make modified versions of the Rast are not
 * obligated to grant this special exception for their modified
 * versions; it is their choice whether to do so.  The GNU General
 * Public License gives permission to release a modified version
 * without this exception; this exception also makes it possible to
 * release a modified version which carries forward this exception.
 */

#ifndef XMLRPC_CLIENT_H
#define XMLRPC_CLIENT_H

/**
 * @file xmlrpc_client.h xmlrpc_client
 */

#include <xmlrpc.h>
#include <xmlrpc_client.h>
#include "rast/db.h"
#include "rast/local_db.h"

/**
 * @defgroup xmlrpc_client xmlrpc_client
 * @{
 */

/** A structure that represents a database to search documents */
typedef struct {
    /** The super class instance */
    rast_db_t base;
    xmlrpc_env env;
    xmlrpc_server_info *server;
    /** The url of the database server*/
    const char *url;
    /** The property definitions */
    rast_property_t *properties;
    /** The number of properties */
    int num_properties;
} rast_xmlrpc_client_t;

rast_error_t *rast_xmlrpc_client_initialize();
void rast_xmlrpc_client_finalize();

/**
 * Open the specified database.
 * @param db The database opened
 * @param name The url of the database
 * @param flags How to open
 *          - RAST_RDRW
 *            - open for reading and writing
 *          - RAST_RDONLY
 *            - open for reading
 * @param options The options to open the database.
 * @param pool The pool to allocate the memory out of
 * @return RAST_OK if succeeded, error otherwise
 * @see rast_db_open(), rast_db_open_option_create()
 */
rast_error_t *rast_xmlrpc_client_open(rast_db_t **db,
                                      const char *name, int flags,
                                      rast_db_open_option_t *options,
                                      apr_pool_t *pool);

/**
 * Please use rast_db_close().
 * @see rast_db_close()
 */
rast_error_t *rast_xmlrpc_client_close(rast_db_t *db);

/**
 * Please use rast_db_register().
 * @see rast_db_register()
 */
rast_error_t *rast_xmlrpc_client_register(rast_db_t *base,
                                          const char *text, rast_size_t nbytes,
                                          rast_value_t *properties,
                                          rast_doc_id_t *new_doc_id);

/**
 * Please use rast_db_search().
 * @see rast_db_search()
 */
rast_error_t *rast_xmlrpc_client_search(rast_db_t *db,
                                        const char *query,
                                        rast_search_option_t *options,
                                        rast_result_t **result,
                                        apr_pool_t *pool);

/**
 * Please use rast_db_delete().
 * @see rast_db_delete()
 */
rast_error_t *rast_xmlrpc_client_delete(rast_db_t *base, rast_doc_id_t doc_id);

/**
 * Please use rast_db_update().
 * @see rast_db_update()
 */
rast_error_t *rast_xmlrpc_client_update(rast_db_t *db, rast_doc_id_t doc_id,
                                        const char *text, rast_size_t nbytes,
                                        rast_value_t *properties,
                                        rast_doc_id_t *new_doc_id);

/**
 * Please use rast_db_get_text().
 * @see rast_db_get_text()
 */
rast_error_t *rast_xmlrpc_client_get_text(rast_db_t *db, rast_doc_id_t doc_id,
                                          char **s, rast_size_t *nbytes,
                                          apr_pool_t *pool);

/**
 * Please use rast_db_properties().
 * @see rast_db_properties()
 */
const rast_property_t *rast_xmlrpc_client_properties(rast_db_t *base,
                                                     int *num_properties);

/** @} */

#endif

/* vim: set filetype=c sw=4 expandtab : */
