/* 
 * Copyright (C) 2005  Network Applied Communication Laboratory Co., Ltd.
 *
 * This file is part of Rast.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA
 *
 * Linking the Rast statically or dynamically with other modules is
 * making a combined work based on the Rast.  Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * As a special exception, the copyright holders of the Rast give
 * you permission to link the Rast with independent modules that
 * communicate with the Rast solely through the the Apache Portable
 * Runtime interface, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * combined work under terms of your choice, provided that every
 * copy of the combined work is accompanied by a complete copy of
 * the source code of the Rast (the version of the Rast used to
 * produce the combined work), being distributed under the terms of
 * the GNU General Public License plus this exception.  An
 * independent module is a module which is not derived from or based
 * on the Rast.
 *
 * Note that people who make modified versions of the Rast are not
 * obligated to grant this special exception for their modified
 * versions; it is their choice whether to do so.  The GNU General
 * Public License gives permission to release a modified version
 * without this exception; this exception also makes it possible to
 * release a modified version which carries forward this exception.
 */

#include <apr_strings.h>
#include <ruby.h>
#include <st.h>

#undef PACKAGE_NAME
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION
#undef PACKAGE_STRING
#include "rast/config.h"
#include "rast/text_index.h"
#include "rast/local_db.h"
#include "rast/query.h"
#include "rast/merger.h"

static VALUE eError, eRastError, eAprError, eBDBError;
static VALUE cPool, cTerm, cTermFrequency, cCandidate, cQueryResult;
static VALUE cResult, cResultTerm, cResultItem;
static VALUE cProperty;

static void
raise_error(rast_error_t *error)
{
    VALUE error_class;
    char buf[RAST_BUFSIZ];

    switch (error->type) {
    case RAST_ERROR_TYPE_RAST:
        error_class = eRastError;
        break;
    case RAST_ERROR_TYPE_APR:
        error_class = eAprError;
        break;
    case RAST_ERROR_TYPE_BDB:
        error_class = eBDBError;
        break;
    default:
        error_class = eError;
    };

#ifdef RAST_DEBUG
    snprintf(buf, RAST_BUFSIZ, "%s:%d: %s",
             error->file, error->line, error->message);
#else
    strncpy(buf, error->message, RAST_BUFSIZ);
#endif
    buf[RAST_BUFSIZ - 1] = '\0';
    rast_error_destroy(error);
    rb_raise(error_class, "%s", buf);
}

static void
pool_free(apr_pool_t *pool)
{
    if (pool) {
        apr_pool_destroy(pool);
    }
}

static VALUE
pool_alloc(VALUE klass)
{
    return Data_Wrap_Struct(klass, NULL, pool_free, NULL);
}

static int
raise_on_pool_failure(int retcode)
{
    rb_memerror();
    return -1;  /* prevent compiler warnings */
}

static VALUE
pool_initialize(VALUE self)
{
    apr_status_t status;
    apr_pool_t *pool;

    status = apr_pool_create_ex(&pool, NULL, raise_on_pool_failure, NULL);
    if (status != APR_SUCCESS) {
        rb_memerror();
    }
    DATA_PTR(self) = pool;
    return Qnil;
}

static VALUE
pool_new()
{
    VALUE obj;

    obj = pool_alloc(cPool);
    pool_initialize(obj);
    return obj;
}

static apr_pool_t *
get_pool(VALUE self)
{
    if (TYPE(self) != T_DATA ||
        RDATA(self)->dfree != (RUBY_DATA_FUNC) pool_free) {
        rb_raise(rb_eTypeError,
                 "wrong argument type %s (expected Rast::Pool)",
                 rb_obj_classname(self));
    }
    return (apr_pool_t *) DATA_PTR(self);
}

static char *
hash_get_string(apr_pool_t *pool, VALUE hash, const char *name)
{
    VALUE val;

    val = rb_hash_aref(hash, rb_str_new2(name));
    SafeStringValue(val);
    return apr_pstrdup(pool, RSTRING(val)->ptr);
}

static int
hash_get_bool(VALUE hash, const char *name)
{
    VALUE val;

    val = rb_hash_aref(hash, rb_str_new2(name));
    return RTEST(val);
}

static rast_type_e
hash_get_property_type(VALUE hash, const char *name)
{
    VALUE val;


    val = rb_hash_aref(hash, rb_str_new2(name));
    return NUM2INT(val);
}

static void
get_bool_option(VALUE options, const char *name, int *dst)
{
    VALUE value = rb_hash_aref(options, rb_str_new2(name));
    if (!NIL_P(value)) {
        *dst = RTEST(value);
    }
}

static void
get_int_option(VALUE options, const char *name, int *dst)
{
    VALUE value = rb_hash_aref(options, rb_str_new2(name));
    if (RTEST(value)) {
        *dst = NUM2INT(value);
    }
}

static void
get_string_option(VALUE options, const char *name, const char **dst)
{
    VALUE value = rb_hash_aref(options, rb_str_new2(name));
    if (RTEST(value)) {
        SafeStringValue(value);
        *dst = RSTRING(value)->ptr;
    }
}

static rast_property_t *
get_properties(apr_pool_t *pool, VALUE vproperties, int *num_properties)
{
    rast_property_t *properties;
    int i;

    Check_Type(vproperties, T_ARRAY);
    properties = (rast_property_t *) apr_palloc(pool,
                                                sizeof(rast_property_t) *
                                                RARRAY(vproperties)->len);
    for (i = 0; i < RARRAY(vproperties)->len; i++) {
        VALUE vproperty = RARRAY(vproperties)->ptr[i];
        rast_property_t *property = properties + i;

        Check_Type(vproperty, T_HASH);
        property->name = hash_get_string(pool, vproperty, "name");
        property->type = hash_get_property_type(vproperty, "type");
        property->flags = 0;
        if (hash_get_bool(vproperty, "search")) {
            property->flags |= RAST_PROPERTY_FLAG_SEARCH;
        }
        if (hash_get_bool(vproperty, "text_search")) {
            property->flags |= RAST_PROPERTY_FLAG_TEXT_SEARCH;
        }
        if (hash_get_bool(vproperty, "full_text_search")) {
            property->flags |= RAST_PROPERTY_FLAG_FULL_TEXT_SEARCH;
        }
        if (hash_get_bool(vproperty, "unique")) {
            property->flags |= RAST_PROPERTY_FLAG_UNIQUE;
        }
        if (hash_get_bool(vproperty, "omit_property")) {
            property->flags |= RAST_PROPERTY_FLAG_OMIT;
        }
    }
    *num_properties = RARRAY(vproperties)->len;
    return properties;
}

static VALUE
process_db_s_create(VALUE self, VALUE vname, VALUE voptions,
                    rast_error_t *(*create)(const char *name,
                                            rast_db_create_option_t *options,
                                            apr_pool_t *pool))
{
    VALUE vpool;
    apr_pool_t *pool;
    VALUE properties;
    rast_error_t *error;
    char *name;
    rast_db_create_option_t *options;

    vpool = pool_new();
    pool = get_pool(vpool);
    SafeStringValue(vname);
    name = RSTRING(vname)->ptr;
    options = rast_db_create_option_create(pool);
    Check_Type(voptions, T_HASH);
    get_int_option(voptions, "byte_order", (int *) &options->byte_order);
    get_int_option(voptions, "pos_block_size", &options->pos_block_size);
    get_string_option(voptions, "encoding", &options->encoding);
    get_bool_option(voptions, "preserve_text", &options->preserve_text);
    properties = rb_hash_aref(voptions, rb_str_new2("properties"));
    options->properties =
        get_properties(pool, properties, &options->num_properties);
    error = create(name, options, pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return Qnil;
}

static VALUE
local_db_s_create(VALUE self, VALUE vname, VALUE voptions)
{
    return process_db_s_create(self, vname, voptions, rast_local_db_create);
}

typedef struct {
    rast_db_t *db;
    apr_pool_t *pool;
    int closed;
} db_data_t;

static void
db_free(db_data_t *data)
{
    if (data) {
        if (!data->closed) {
            rast_db_close(data->db);
        }
        apr_pool_destroy(data->pool);
        xfree(data);
    }
}

static VALUE
db_alloc(VALUE klass)
{
    return Data_Wrap_Struct(klass, NULL, db_free, NULL);
}

static VALUE
process_db_initialize(int argc, VALUE *argv, VALUE self,
                      rast_error_t *(*open)(rast_db_t **db, const char *name,
                                            int flags,
                                            rast_db_open_option_t *options,
                                            apr_pool_t *pool))
{
    VALUE name, vflags, voptions, vpool;
    rast_error_t *error;
    apr_pool_t *pool, *tmp_pool;
    db_data_t *data;
    rast_db_t *db;
    int flags = RAST_DB_RDWR;
    rast_db_open_option_t *options = NULL;

    rb_scan_args(argc, argv, "12", &name, &vflags, &voptions);
    if (!NIL_P(vflags)) {
        flags = NUM2INT(vflags);
    }
    if (!NIL_P(voptions)) {
        vpool = pool_new();
        tmp_pool = get_pool(vpool);
        options = rast_db_open_option_create(tmp_pool);
        get_int_option(voptions, "sync_threshold_chars",
                       &options->sync_threshold_chars);
    }
    apr_pool_create_ex(&pool, NULL, raise_on_pool_failure, NULL);
    SafeStringValue(name);
    error = open(&db, RSTRING(name)->ptr, flags, options, pool);
    if (error != RAST_OK) {
        apr_pool_destroy(pool);
        raise_error(error);
    }
    data = ALLOC(db_data_t);
    data->db = db;
    data->pool = pool;
    data->closed = 0;
    DATA_PTR(self) = data;
    return Qnil;
}

static VALUE
db_initialize(int argc, VALUE *argv, VALUE self)
{
    return process_db_initialize(argc, argv, self, rast_db_open);
}

static VALUE
local_db_initialize(int argc, VALUE *argv, VALUE self)
{
    return process_db_initialize(argc, argv, self, rast_local_db_open);
}

static rast_db_t *
get_db(VALUE self)
{
    db_data_t *data;

    if (TYPE(self) != T_DATA ||
        RDATA(self)->dfree != (RUBY_DATA_FUNC) db_free) {
        rb_raise(rb_eTypeError,
                 "wrong argument type %s (expected Rast::DB)",
                 rb_obj_classname(self));
    }

    data = DATA_PTR(self);
    if (data->closed) {
        rb_raise(eError, "already closed db");
    }
    return data->db;
}

static VALUE
db_close(VALUE self)
{
    db_data_t *data;
    rast_error_t *error;

    Data_Get_Struct(self, db_data_t, data);
    error = rast_db_close(data->db);
    if (error != RAST_OK) {
        raise_error(error);
    }
    data->closed = 1;
    return Qnil;
}

static VALUE
db_s_create(VALUE self, VALUE vname, VALUE voptions)
{
    return process_db_s_create(self, vname, voptions, rast_db_create);
}

static VALUE
db_s_open(int argc, VALUE *argv, VALUE self)
{
    VALUE db = rb_class_new_instance(argc, argv, self);
    if (rb_block_given_p()) {
        return rb_ensure(rb_yield, db, db_close, db);
    }

    return db;
}

static VALUE
db_byte_order(VALUE self)
{
    rast_db_t *db;
    rast_byte_order_e byte_order;

    db = get_db(self);
    byte_order = rast_db_byte_order(db);
    if (byte_order == RAST_UNKNOWN_ENDIAN) {
        rb_raise(eRastError, "unknown endian");
    }
    return INT2NUM(byte_order);
}

static VALUE
db_encoding(VALUE self)
{
    rast_db_t *db;
    const char *encoding;

    db = get_db(self);
    encoding = rast_db_encoding(db);
    if (encoding == NULL) {
        rb_raise(eRastError, "unknown encoding");
    }
    return rb_tainted_str_new2(encoding);
}

static VALUE
db_properties(VALUE self)
{
    rast_db_t *db;
    const rast_property_t *properties;
    VALUE result;
    int i, num_properties;

    db = get_db(self);
    properties = rast_db_properties(db, &num_properties);
    result = rb_ary_new();
    for (i = 0; i < num_properties; i++) {
        const rast_property_t *property = properties + i;
        VALUE val, search, text_search, full_text_search, unique;
        VALUE omit_property;

        search = property->flags & RAST_PROPERTY_FLAG_SEARCH ? Qtrue : Qfalse;
        text_search =
            property->flags & RAST_PROPERTY_FLAG_TEXT_SEARCH ? Qtrue : Qfalse;
        full_text_search =
            property->flags & RAST_PROPERTY_FLAG_FULL_TEXT_SEARCH ?
            Qtrue : Qfalse;
        unique = property->flags & RAST_PROPERTY_FLAG_UNIQUE ? Qtrue : Qfalse;
        omit_property = property->flags & RAST_PROPERTY_FLAG_OMIT ?
            Qtrue : Qfalse;
        val = rb_funcall(cProperty, rb_intern("new"), 7,
                         rb_tainted_str_new2(property->name),
                         INT2NUM(property->type),
                         search, text_search, full_text_search, unique,
                         omit_property);
        rb_ary_push(result, val);
    }
    return result;
}

static VALUE
db_sync_threshold_chars(VALUE self)
{
    rast_db_t *db = get_db(self);
    int sync_threshold_chars;

    sync_threshold_chars = rast_db_sync_threshold_chars(db);
    if (sync_threshold_chars == -1) {
        rb_raise(eRastError, "unknown sync_threshold_chars");
    }
    return INT2NUM(sync_threshold_chars);
}

static rast_value_t *
get_property_values(VALUE vproperty_values, rast_db_t *db, apr_pool_t *pool)
{
    const rast_property_t *properties;
    rast_value_t *property_values;
    int i, num_properties;

    properties = rast_db_properties(db, &num_properties);
    property_values = (rast_value_t *)
        apr_palloc(pool, sizeof(rast_value_t) * num_properties);
    for (i = 0; i < num_properties; i++) {
        const rast_property_t *property = properties + i;
        VALUE value;

        value = rb_hash_aref(vproperty_values, rb_str_new2(property->name));
        switch (property->type) {
        case RAST_TYPE_STRING:
        case RAST_TYPE_DATE:
            Check_Type(value, T_STRING);
            rast_value_set_string(property_values + i, StringValuePtr(value));
            break;
        case RAST_TYPE_UINT:
            Check_Type(value, T_FIXNUM);
            rast_value_set_uint(property_values + i, NUM2INT(value));
            break;
        default:
            rb_raise(eError, "unknown property type");
        };
    }
    return property_values;
}

static VALUE
db_register(VALUE self, VALUE text, VALUE vproperty_values)
{
    rast_db_t *db;
    rast_value_t *property_values;
    rast_error_t *error;
    apr_pool_t *pool;
    VALUE vpool = pool_new();
    rast_doc_id_t doc_id;

    pool = get_pool(vpool);
    db = get_db(self);
    property_values = get_property_values(vproperty_values, db, pool);
    StringValue(text);
    error = rast_db_register(db, RSTRING(text)->ptr, RSTRING(text)->len,
                             property_values, &doc_id);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return INT2NUM(doc_id);
}

static VALUE
result_new(rast_result_t *result, int num_properties)
{
    VALUE terms, term, items, vitem, properties;
    int i, j;

    terms = rb_ary_new();
    for (i = 0; i < result->num_terms; i++) {
        term = rb_funcall(cResultTerm, rb_intern("new"), 2,
                          rb_tainted_str_new2(result->terms[i].term),
                          INT2NUM(result->terms[i].doc_count));
        rb_ary_push(terms, term);
    }

    items = rb_ary_new();
    for (i = 0; i < result->num_items; i++) {
        rast_result_item_t *item = result->items[i];

        properties = rb_ary_new();
        for (j = 0; j < num_properties; j++) {
            rast_value_t *value = item->properties + j;
            switch (value->type) {
            case RAST_TYPE_STRING:
            case RAST_TYPE_DATE:
                rb_ary_push(properties,
                            rb_tainted_str_new2(rast_value_string(value)));
                break;
            case RAST_TYPE_UINT:
                rb_ary_push(properties, INT2NUM(rast_value_uint(value)));
                break;
            }
        }
        vitem = rb_funcall(cResultItem, rb_intern("new"), 5,
                           INT2NUM(item->doc_id),
                           INT2NUM(item->db_index),
                           INT2NUM(item->score),
                           rb_tainted_str_new(item->summary,
                                              item->summary_nbytes),
                           properties);
        rb_ary_push(items, vitem);
    }
    return rb_funcall(cResult, rb_intern("new"), 5,
                      INT2NUM(result->num_indices),
                      INT2NUM(result->num_docs),
                      INT2NUM(result->hit_count), terms, items);
}

static VALUE
db_sync(VALUE self)
{
    rast_db_t *db;
    rast_error_t *error;

    db = get_db(self);
    error = rast_db_sync(db);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return Qnil;
}

static VALUE
db_search(int argc, VALUE *argv, VALUE self)
{
    VALUE query, voptions, properties;
    rast_db_t *db;
    rast_error_t *error;
    rast_search_option_t *options;
    rast_result_t *result;
    int i;
    apr_pool_t *pool;
    VALUE vpool = pool_new(), value;

    pool = get_pool(vpool);
    db = get_db(self);
    options = rast_search_option_create(pool);
    if (rb_scan_args(argc, argv, "11", &query, &voptions) == 2) {
        VALUE terms;

        Check_Type(voptions, T_HASH);

        get_int_option(voptions, "start_no", &options->start_no);
        get_int_option(voptions, "num_items", &options->num_items);
        options->need_summary = hash_get_bool(voptions, "need_summary");
        get_int_option(voptions, "summary_nchars", &options->summary_nchars);
        get_int_option(voptions, "sort_order", (int *) &options->sort_order);
        value = rb_hash_aref(voptions, rb_str_new2("sort_property"));
        if (!NIL_P(value)) {
            options->sort_property = apr_pstrdup(pool, StringValuePtr(value));
        }
        get_int_option(voptions, "sort_method", (int *) &options->sort_method);
        get_int_option(voptions, "score_method",
                       (int *) &options->score_method);
        get_int_option(voptions, "all_num_docs",
                       (int *) &options->all_num_docs);
        terms = rb_hash_aref(voptions, rb_str_new2("terms"));
        if (!NIL_P(terms)) {
            Check_Type(terms, T_ARRAY);
            options->num_terms = RARRAY(terms)->len;
            options->terms = (int *)
                apr_palloc(pool, sizeof(int) * options->num_terms);
            for (i = 0; i < options->num_terms; i++) {
                options->terms[i] = NUM2INT(RARRAY(terms)->ptr[i]);
            }
        }
        properties = rb_hash_aref(voptions, rb_str_new2("properties"));
        if (!NIL_P(properties)) {
            int properties_len = sizeof(char *) * RARRAY(properties)->len;

            Check_Type(properties, T_ARRAY);
            options->properties =
                (const char **) apr_palloc(pool, properties_len);
            for (i = 0; i < RARRAY(properties)->len; i++) {
                options->properties[i] =
                    StringValuePtr(RARRAY(properties)->ptr[i]);
            }
            options->num_properties = RARRAY(properties)->len;
        }
    }
    error = rast_db_search(db, StringValuePtr(query), options, &result, pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return result_new(result, options->num_properties);
}

static VALUE
db_delete(VALUE self, VALUE vdoc_id)
{
    rast_doc_id_t doc_id;
    rast_db_t *db;
    rast_error_t *error;

    doc_id = (rast_doc_id_t) NUM2INT(vdoc_id);
    db = get_db(self);
    error = rast_db_delete(db, doc_id);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return Qnil;
}

static VALUE
db_update(VALUE self, VALUE vdoc_id, VALUE text, VALUE vproperty_values)
{
    rast_db_t *db;
    rast_doc_id_t doc_id, new_doc_id;
    rast_value_t *property_values;
    rast_error_t *error;
    apr_pool_t *pool;
    VALUE vpool = pool_new();

    doc_id = (rast_doc_id_t) NUM2INT(vdoc_id);
    pool = get_pool(vpool);
    db = get_db(self);
    property_values = get_property_values(vproperty_values, db, pool);
    StringValue(text);
    error = rast_db_update(db, doc_id, RSTRING(text)->ptr, RSTRING(text)->len,
                           property_values, &new_doc_id);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return INT2NUM(new_doc_id);
}

static VALUE
db_get_text(VALUE self, VALUE doc_id)
{
    rast_db_t *db;
    rast_error_t *error;
    apr_pool_t *pool;
    VALUE vpool = pool_new();
    char *s;
    rast_size_t len;

    pool = get_pool(vpool);
    db = get_db(self);
    error = rast_db_get_text(db, NUM2INT(doc_id), &s, &len, pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return rb_tainted_str_new(s, len);
}

static VALUE
merger_initialize(int argc, VALUE *argv, VALUE self)
{
    VALUE vdbs;
    rast_error_t *error;
    apr_pool_t *pool;
    db_data_t *data;
    rast_db_t *db;
    int i, num_dbs;
    rast_db_t **dbs;

    rb_scan_args(argc, argv, "10", &vdbs);

    apr_pool_create_ex(&pool, NULL, raise_on_pool_failure, NULL);

    Check_Type(vdbs, T_ARRAY);
    num_dbs = RARRAY(vdbs)->len;
    dbs = (rast_db_t **)
        apr_palloc(pool, sizeof(rast_db_t *) * num_dbs);
    for (i = 0; i < num_dbs; i++) {
        dbs[i] = get_db(RARRAY(vdbs)->ptr[i]);
    }

    error = rast_merger_open(&db, dbs, num_dbs, pool);
    if (error != RAST_OK) {
        apr_pool_destroy(pool);
        raise_error(error);
    }

    data = ALLOC(db_data_t);
    data->db = db;
    data->pool = pool;
    data->closed = 0;
    DATA_PTR(self) = data;
    return Qnil;
}

void
Init_rast()
{
    VALUE mRast, cDB, cLocalDB, cMerger;

    apr_initialize();
    atexit(apr_terminate);

    rast_initialize();
    atexit(rast_finalize);

    mRast = rb_define_module("Rast");

    eError = rb_define_class_under(mRast, "Error", rb_eStandardError);
    rb_define_const(eError, "TYPE_RAST", INT2NUM(RAST_ERROR_TYPE_RAST));
    rb_define_const(eError, "TYPE_APR", INT2NUM(RAST_ERROR_TYPE_APR));
    rb_define_const(eError, "TYPE_BDB", INT2NUM(RAST_ERROR_TYPE_BDB));
    eRastError = rb_define_class_under(mRast, "RastError", eError);
    eAprError = rb_define_class_under(mRast, "AprError", eError);
    eBDBError = rb_define_class_under(mRast, "BDBError", eError);

    rb_define_const(mRast, "VERSION", rb_str_new2(VERSION));

    rb_define_const(mRast, "RESULT_ALL_ITEMS", INT2NUM(RAST_RESULT_ALL_ITEMS));

    rb_define_const(mRast, "NATIVE_ENDIAN", INT2NUM(RAST_NATIVE_ENDIAN));
    rb_define_const(mRast, "LITTLE_ENDIAN", INT2NUM(RAST_LITTLE_ENDIAN));
    rb_define_const(mRast, "BIG_ENDIAN", INT2NUM(RAST_BIG_ENDIAN));

    cPool = rb_define_class_under(mRast, "Pool", rb_cObject);
    rb_define_alloc_func(cPool, pool_alloc);
    rb_define_method(cPool, "initialize", pool_initialize, 0);

    cTerm = rb_funcall(rb_cStruct, rb_intern("new"), 2,
                       ID2SYM(rb_intern("term")),
                       ID2SYM(rb_intern("doc_count")));
    rb_define_const(mRast, "Term", cTerm);
    cTermFrequency = rb_funcall(rb_cStruct, rb_intern("new"), 2,
                                ID2SYM(rb_intern("count")),
                                ID2SYM(rb_intern("pos")));
    rb_define_const(mRast, "TermFrequency", cTermFrequency);
    cCandidate = rb_funcall(rb_cStruct, rb_intern("new"), 2,
                            ID2SYM(rb_intern("doc_id")),
                            ID2SYM(rb_intern("terms")));
    rb_define_const(mRast, "Candidate", cCandidate);
    cQueryResult = rb_funcall(rb_cStruct, rb_intern("new"), 2,
                              ID2SYM(rb_intern("terms")),
                              ID2SYM(rb_intern("candidates")));
    rb_define_const(mRast, "QueryResult", cQueryResult);
    cResultTerm = rb_funcall(rb_cStruct, rb_intern("new"), 2,
                             ID2SYM(rb_intern("term")),
                             ID2SYM(rb_intern("doc_count")));
    rb_define_const(mRast, "ResultTerm", cResultTerm);
    cResultItem = rb_funcall(rb_cStruct, rb_intern("new"), 5,
                             ID2SYM(rb_intern("doc_id")),
                             ID2SYM(rb_intern("db_index")),
                             ID2SYM(rb_intern("score")),
                             ID2SYM(rb_intern("summary")),
                             ID2SYM(rb_intern("properties")));
    rb_define_const(mRast, "ResultItem", cResultItem);
    cResult = rb_funcall(rb_cStruct, rb_intern("new"), 5,
                         ID2SYM(rb_intern("num_indices")),
                         ID2SYM(rb_intern("num_docs")),
                         ID2SYM(rb_intern("hit_count")),
                         ID2SYM(rb_intern("terms")),
                         ID2SYM(rb_intern("items")));
    rb_define_const(mRast, "Result", cResult);

    cProperty = rb_funcall(rb_cStruct, rb_intern("new"), 7,
                           ID2SYM(rb_intern("name")),
                           ID2SYM(rb_intern("type")),
                           ID2SYM(rb_intern("search")),
                           ID2SYM(rb_intern("text_search")),
                           ID2SYM(rb_intern("full_text_search")),
                           ID2SYM(rb_intern("unique")),
                           ID2SYM(rb_intern("omit_property")));
    rb_define_const(mRast, "Property", cProperty);

    rb_define_const(mRast, "PROPERTY_TYPE_STRING", INT2NUM(RAST_TYPE_STRING));
    rb_define_const(mRast, "PROPERTY_TYPE_UINT", INT2NUM(RAST_TYPE_UINT));
    rb_define_const(mRast, "PROPERTY_TYPE_DATE", INT2NUM(RAST_TYPE_DATE));

    rb_define_const(mRast, "SORT_METHOD_SCORE",
                    INT2NUM(RAST_SORT_METHOD_SCORE));
    rb_define_const(mRast, "SORT_METHOD_PROPERTY",
                    INT2NUM(RAST_SORT_METHOD_PROPERTY));

    rb_define_const(mRast, "SORT_ORDER_DEFAULT",
                    INT2NUM(RAST_SORT_ORDER_DEFAULT));
    rb_define_const(mRast, "SORT_ORDER_ASCENDING",
                    INT2NUM(RAST_SORT_ORDER_ASCENDING));
    rb_define_const(mRast, "SORT_ORDER_DESCENDING",
                    INT2NUM(RAST_SORT_ORDER_DESCENDING));

    rb_define_const(mRast, "SCORE_METHOD_NONE",
                    INT2NUM(RAST_SCORE_METHOD_NONE));
    rb_define_const(mRast, "SCORE_METHOD_TFIDF",
                    INT2NUM(RAST_SCORE_METHOD_TFIDF));

    cDB = rb_define_class_under(mRast, "DB", rb_cObject);
    rb_define_alloc_func(cDB, db_alloc);
    rb_define_const(cDB, "RDWR", INT2NUM(RAST_DB_RDWR));
    rb_define_const(cDB, "RDONLY", INT2NUM(RAST_DB_RDONLY));
    rb_define_singleton_method(cDB, "create", db_s_create, 2);
    rb_define_singleton_method(cDB, "open", db_s_open, -1);
    rb_define_method(cDB, "initialize", db_initialize, -1);
    rb_define_method(cDB, "sync", db_sync, 0);
    rb_define_method(cDB, "close", db_close, 0);
    rb_define_method(cDB, "register", db_register, 2);
    rb_define_method(cDB, "search", db_search, -1);
    rb_define_method(cDB, "delete", db_delete, 1);
    rb_define_method(cDB, "update", db_update, 3);
    rb_define_method(cDB, "get_text", db_get_text, 1);
    rb_define_method(cDB, "byte_order", db_byte_order, 0);
    rb_define_method(cDB, "encoding", db_encoding, 0);
    rb_define_method(cDB, "properties", db_properties, 0);
    rb_define_method(cDB, "sync_threshold_chars", db_sync_threshold_chars, 0);

    cLocalDB = rb_define_class_under(mRast, "LocalDB", cDB);
    rb_define_singleton_method(cLocalDB, "create", local_db_s_create, 2);
    rb_define_method(cLocalDB, "initialize", local_db_initialize, -1);

    cMerger = rb_define_class_under(mRast, "Merger", cDB);
    rb_define_method(cMerger, "initialize", merger_initialize, -1);
}

/* vim: set filetype=c sw=4 expandtab : */
