# Copyright (C) 2005  Network Applied Communication Laboratory Co., Ltd.
#
# This file is part of Rast.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA
#
# Linking the Rast statically or dynamically with other modules is
# making a combined work based on the Rast.  Thus, the terms and
# conditions of the GNU General Public License cover the whole
# combination.
#
# As a special exception, the copyright holders of the Rast give
# you permission to link the Rast with independent modules that
# communicate with the Rast solely through the the Apache Portable
# Runtime interface, regardless of the license terms of these
# independent modules, and to copy and distribute the resulting
# combined work under terms of your choice, provided that every
# copy of the combined work is accompanied by a complete copy of
# the source code of the Rast (the version of the Rast used to
# produce the combined work), being distributed under the terms of
# the GNU General Public License plus this exception.  An
# independent module is a module which is not derived from or based
# on the Rast.
#
# Note that people who make modified versions of the Rast are not
# obligated to grant this special exception for their modified
# versions; it is their choice whether to do so.  The GNU General
# Public License gives permission to release a modified version
# without this exception; this exception also makes it possible to
# release a modified version which carries forward this exception.

require "test/unit"
require "socket"

require "rast/xmlrpc-server-testable"

require "rast_test"
require "rast_xmlrpc_client"

module Rast
  class XMLRPCClientTest < Test::Unit::TestCase
    include XMLRPCServerTestable

    def test_register_raw
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "title",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "modified",
            "type" => Rast::PROPERTY_TYPE_DATE,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "serial",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
        ],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      kick_xmlrpc_client("--read-write", db_name) do |db|
        properties = {
          "title" => "test",
          "modified" => "2005-02-17T19:04:38",
          "serial" => 10,
        }
        assert_equal(1, db.register_raw("本日は晴天なり",
                                        [
                                          "test",
                                          DateTime.new(2005, 4, 15),
                                          10
                                        ]))
      end
    end

    def test_register
      register_test_simple
      register_test_only_property
    end

    def register_test_simple
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      kick_xmlrpc_client("--read-write", db_name) do |db|
        assert_equal(1, db.register("本日は晴天なり", {}))
      end
      LocalDB.open(db_name) do |db|
        result = db.search("晴天")
        assert_equal(1, result.num_docs)
        assert_equal(1, result.hit_count)
        assert_equal("晴天", result.terms[0].term)
        assert_equal(1, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(1, result.items[0].doc_id)
        assert_equal(0, result.items[0].properties.length)
        assert_equal(1, result.items.length)
      end
    end

    def register_test_only_property
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "title",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "modified",
            "type" => Rast::PROPERTY_TYPE_DATE,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "serial",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
          },
        ],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      kick_xmlrpc_client("--read-write", db_name) do |db|
        properties = {
          "title" =>"test",
          "modified" => "2005-02-17T19:04:38",
          "serial" => 10,
        }
        assert_equal(1, db.register("", properties))
      end
      LocalDB.open(db_name) do |db|
        result = db.search("serial = 10",
                           {
                             "properties" => [
                               "title",
                               "modified",
                               "serial",
                             ]
                           })
        assert_equal(1, result.items[0].doc_id)
        assert_equal("test", result.items[0].properties[0])
        assert_equal("2005-02-17T19:04:38", result.items[0].properties[1])
        assert_equal(10, result.items[0].properties[2])
        assert_equal(3, result.items[0].properties.length)
      end
    end

    def test_search
      search_test_simple
      search_test_summary
      search_test_sort_with_score
      search_test_calc_score
      search_test_calc_received_all_num_docs_score
      search_test_sort_with_property
      search_test_set_range
      search_test_rast_error
    end

    def search_test_simple
      options = {
        "encoding" => "utf8",
        "properties" => [],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("昨日は雨天なり", {})
        db.register("本日は晴天なり", {})
        db.register("明日は雪っすり", {})
      end
      kick_xmlrpc_client(db_name) do |db|
        result = db.search("晴天")
        assert_equal(1, result.num_indices)
        assert_equal(3, result.num_docs)
        assert_equal(1, result.hit_count)
        assert_equal("晴天", result.terms[0].term)
        assert_equal(1, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal(0, result.items[0].db_index)
        assert_equal(0, result.items[0].properties.length)
        assert_equal(1, result.items.length)

        result = db.search("なり")
        assert_equal(2, result.hit_count)
        assert_equal("なり", result.terms[0].term)
        assert_equal(2, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(1, result.items[0].doc_id)
        assert_equal(0, result.items[0].properties.length)
        assert_equal(2, result.items[1].doc_id)
        assert_equal(0, result.items[1].properties.length)
        assert_equal(2, result.items.length)
      end
    end

    def search_test_summary
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "title",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => true,
            "unique" => false,
          },
        ],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("0123456789", {"title" => "full text search property"})
        db.register("零一二三四五六七八九",
                    {"title" => "full text search property"})
        db.register("零1二3四5六7八9",
                    {"title" => "full text search property"})
      end
      kick_xmlrpc_client(db_name) do |db|
        search_options = {"need_summary" => true, "summary_nchars" => 6}
        result = db.search("5", search_options)
        assert_equal(1, result.items[0].doc_id)
        assert_equal("345678", result.items[0].summary)
        assert_equal(3, result.items[1].doc_id)
        assert_equal("3四5六7八", result.items[1].summary)
        assert_equal(2, result.items.length)
      end
    end

    def search_test_sort_with_score
      options = {
        "encoding" => "utf8",
        "properties" => [],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("あい" * 99 + "晴天", {}) 
        db.register("晴天" * 100, {})
        db.register("あい" * 49 + "晴天", {}) 
      end
      kick_xmlrpc_client(db_name) do |db|
        # descend
        result = db.search("晴天")
        assert_equal(3, result.hit_count)
        assert_equal("晴天", result.terms[0].term)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal(3, result.items[1].doc_id)
        assert_equal(1, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        # ascend
        search_options = {
          "sort_order" => Rast::SORT_ORDER_ASCENDING
        }
        result = db.search("晴天", search_options)
        assert_equal(3, result.hit_count)
        assert_equal("晴天", result.terms[0].term)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(1, result.items[0].doc_id)
        assert_equal(3, result.items[1].doc_id)
        assert_equal(2, result.items[2].doc_id)
        assert_equal(3, result.items.length)
      end
    end

    def search_test_calc_score
      options = {
        "encoding" => "utf8",
        "properties" => [
          {
            "name" => "filename",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
          },
        ],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("aaa" * 100, {"filename" => "abc.txt"})
      end
      kick_xmlrpc_client(db_name) do |db|
        options = {
          "score_method" => Rast::SCORE_METHOD_NONE,
        }
        result = db.search("aaa", options)
        assert_equal(0, result.items[0].score)

        options = {
          "score_method" => Rast::SCORE_METHOD_TFIDF,
        }
        result = db.search("aaa", options)
        assert_operator(0, :<, result.items[0].score)

        result = db.search("aaa", {})
        assert_operator(0, :<, result.items[0].score)

        options = {
          "sort_method" => Rast::SORT_METHOD_PROPERTY,
          "sort_property" => "filename"
        }
        result = db.search("aaa", options)
        assert_operator(0, :<, result.items[0].score)

        options = {
          "score_method" => Rast::SCORE_METHOD_TFIDF,
          "sort_method" => Rast::SORT_METHOD_PROPERTY,
          "sort_property" => "filename"
        }
        result = db.search("aaa", options)
        assert_operator(0, :<, result.items[0].score)
      end
    end

    def search_test_calc_received_all_num_docs_score
      options = {
        "encoding" => "utf8",
        "properties" => [
        ],
      }
      db_name1 = generate_db_name
      LocalDB.create(db_name1, options)
      LocalDB.open(db_name1) do |db|
        db.register("qwerty abcdef", {})
        db.register("qwerty", {})
      end
      result1 = LocalDB.open(db_name1) do |db|
        db.search("qwe abc", {})
      end
      db_name2 = generate_db_name
      LocalDB.create(db_name2, options)
      LocalDB.open(db_name2) do |db|
        db.register("qwerty abcdef", {})
      end
      kick_xmlrpc_client(db_name2) do |db|
        terms = result1.terms.collect do |term|
          term.doc_count
        end
        result2 = db.search("qwe abc",
                            {"all_num_docs" => 2, "terms" => terms})
        assert_equal(result1.items[0].score, result2.items[0].score)
      end
    end

    def search_test_sort_with_property
      options = {
        "encoding" => "utf8",
        "properties" => [
          {
            "name" => "filename",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
          {
            "name" => "serial",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => false,
            "text_search" => false,
            "full_text_search" => false,
          },
        ],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("昨日は雨天なり",
                    {"filename" => "2.txt", "serial" => 10})
        db.register("本日は晴天なり",
                    {"filename" => "3.txt", "serial" => 5})
        db.register("明日は雪っすり",
                    {"filename" => "1.txt", "serial" => 15})
      end
      kick_xmlrpc_client(db_name) do |db|
        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "filename",
                             "properties" => ["filename"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(3, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(2, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "filename",
                             "sort_order" => Rast::SORT_ORDER_ASCENDING,
                             "properties" => ["filename"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(3, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(2, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "filename",
                             "sort_order" => Rast::SORT_ORDER_DESCENDING,
                             "properties" => ["filename"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(3, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "serial",
                             "properties" => ["serial"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(3, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "serial",
                             "sort_order" => Rast::SORT_ORDER_ASCENDING,
                             "properties" => ["serial"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(2, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(3, result.items[2].doc_id)
        assert_equal(3, result.items.length)

        result = db.search("日は",
                           {
                             "sort_method" => Rast::SORT_METHOD_PROPERTY,
                             "sort_property" => "serial",
                             "sort_order" => Rast::SORT_ORDER_DESCENDING,
                             "properties" => ["serial"],
                           })
        assert_equal(3, result.hit_count)
        assert_equal(3, result.terms[0].doc_count)
        assert_equal(1, result.terms.length)
        assert_equal(3, result.items[0].doc_id)
        assert_equal(1, result.items[1].doc_id)
        assert_equal(2, result.items[2].doc_id)
        assert_equal(3, result.items.length)
      end
    end

    def search_test_set_range
      options = {
        "encoding" => "utf8",
        "properties" => [],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("1", {}) 
        db.register("1", {}) 
        db.register("1", {}) 
        db.register("1", {}) 
        db.register("1", {}) 
      end
      kick_xmlrpc_client(db_name) do |db|
        search_options = {
          "start_no" => 0,
          "num_items" => 3,
        }
        result = db.search("1", search_options)
        assert_equal(5, result.hit_count)
        assert_equal(3, result.items.length)

        search_options = {
          "start_no" => 1,
          "num_items" => 3,
        }
        result = db.search("1", search_options)
        assert_equal(5, result.hit_count)
        assert_equal(3, result.items.length)

        search_options = {
          "start_no" => 4,
          "num_items" => 10,
        }
        result = db.search("1", search_options)
        assert_equal(5, result.hit_count)
        assert_equal(1, result.items.length)

        search_options = {
          "start_no" => 20,
          "num_items" => 10,
        }
        result = db.search("1", search_options)
        assert_equal(5, result.hit_count)
        assert_equal(0, result.items.length)
      end
    end

    def search_test_rast_error
      options = {
        "encoding" => "utf8",
        "properties" => [],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("昨日は雨天なり", {})
      end
      kick_xmlrpc_client(db_name) do |db|
        assert_raise(RastError) do
          db.search("晴天", {"properties" => ["invalid_property"]})
        end
      end
    end

    def test_delete
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("本日は晴天なり", {})
      end
      kick_xmlrpc_client("--read-write", db_name) do |db|
        result = db.search("晴天")
        assert_equal(1, result.hit_count)
        db.delete(1)
      end
      LocalDB.open(db_name) do |db|
        result = db.search("晴天")
        assert_equal(0, result.hit_count)
      end
    end

    def test_update
      update_test_simple
      update_test_property
    end

    def update_test_simple
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("本日は晴天なり", {})
      end
      kick_xmlrpc_client("--read-write", db_name) do |db|
        result = db.search("晴天")
        assert_equal(1, result.hit_count)
        assert_equal(2, db.update(1, "明日は雪っす", {}))
      end
      LocalDB.open(db_name) do |db|
        result = db.search("晴天")
        assert_equal(0, result.hit_count)
        result = db.search("雪っ")
        assert_equal(1, result.hit_count)
      end
    end

    def update_test_property
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [
          {
            "name" => "filename",
            "type" => Rast::PROPERTY_TYPE_STRING,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
            "unique" => true,
          },
          {
            "name" => "count",
            "type" => Rast::PROPERTY_TYPE_UINT,
            "search" => true,
            "text_search" => false,
            "full_text_search" => false,
            "unique" => false,
          },
        ],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("本日は晴天なり", {"filename" => "a.txt", "count" => 10})
      end
      LocalDB.open(db_name) do |db|
        result = db.search("晴天")
        assert_equal(1, result.hit_count)
      end
      kick_xmlrpc_client("--read-write", db_name) do |db|
        assert_equal(2, db.update(1, "明日は雪っす",
                                  {"filename" => "b.txt", "count" => 20}))
      end
      LocalDB.open(db_name) do |db|
        result = db.search("晴天")
        assert_equal(0, result.hit_count)
        result = db.search("雪っ")
        assert_equal(1, result.hit_count)
        result = db.search("count = 10")
        assert_equal(0, result.hit_count)
        result = db.search("count = 20")
        assert_equal(1, result.hit_count)
      end
    end

    def test_get_text
      options = {
        "encoding" => "utf8",
        "preserve_text" => true,
        "properties" => [],
      }
      db_name = generate_db_name
      text = "本日は晴天なり"
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register(text, {})
      end
      kick_xmlrpc_client(db_name) do |db|
        assert_equal(text, db.get_text(1))
      end

      options = {
        "encoding" => "utf8",
        "preserve_text" => false,
        "properties" => [],
      }
      db_name = generate_db_name
      LocalDB.create(db_name, options)
      LocalDB.open(db_name) do |db|
        db.register("本日は晴天なり", {})
      end
      LocalDB.open(db_name) do |db|
        assert_raise(RastError) do
          db.get_text(1)
        end
      end
    end
  end
end
